package djmil.cordacheckers.gamestate;

import java.util.UUID;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.states.GameState;
import net.corda.v5.application.flows.ClientRequestBody;
import net.corda.v5.application.flows.ClientStartableFlow;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.FlowEngine;
import net.corda.v5.application.flows.SubFlow;
import net.corda.v5.application.marshalling.JsonMarshallingService;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.crypto.SecureHash;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.UtxoLedgerService;

public class GetFlow implements ClientStartableFlow, SubFlow<StateAndRef<GameState>> {

    private final static Logger log = LoggerFactory.getLogger(GetFlow.class);
    private UUID subflowGameStateUuid = null;

    public GetFlow() {
        this.subflowGameStateUuid = null;
    }

    public GetFlow(UUID gameStateUuid) {
        this.subflowGameStateUuid = gameStateUuid;
    }

    @CordaInject
    public JsonMarshallingService jsonMarshallingService;

    @CordaInject
    public UtxoLedgerService utxoLedgerService;

    @CordaInject
    public FlowEngine flowEngine;

    @Override
    @Suspendable
    public StateAndRef<GameState> call() {              // <<-- SubFlow<T> entry
        return findGameStateUtxoSar(this.subflowGameStateUuid);
    }

    @Override
    @Suspendable
    public String call(ClientRequestBody requestBody) { // <<-- ClientStartableFlow entry
        SecureHash gameStateUtxoTrxId = null;

        try {
            final UUID gameStateUuid = UUID.fromString(requestBody.getRequestBody());

            gameStateUtxoTrxId = findGameStateUtxoSar(gameStateUuid)
                .getRef().getTransactionId();

            final View gameStateView = this.flowEngine
                .subFlow(new ViewBuilder(gameStateUtxoTrxId));

            return new FlowResponce(gameStateView, gameStateUtxoTrxId)
                .toJsonEncodedString(jsonMarshallingService);
        }
        catch (Exception e) {
            log.warn("Exception during processing " + requestBody + " request: " + e.getMessage());
            return new FlowResponce(e, gameStateUtxoTrxId).toJsonEncodedString(jsonMarshallingService);
        }
    }

    @Suspendable
    StateAndRef<GameState> findGameStateUtxoSar(UUID gameStateUuid) {
        /*
         * Get list of all unconsumed aka 'active' GameStates, then filter by UUID.
         * Note, this is an inefficient way to perform this operation if there are a large
         * number of 'active' GameProposals exists in storage.
         */
        return this.utxoLedgerService
            .findUnconsumedStatesByType(GameState.class)
            .stream()
            .filter(sar -> sar.getState().getContractState().getGameUuid().equals(gameStateUuid))
            .reduce((a, b) -> {throw new IllegalStateException("Multiple GameStates found: " +a +", " +b);})
            .get();
    }

}
